@@ -1173,6 +1173,8 @@ static void close_sco(struct audio_device *device)
 	struct headset *hs = device->headset;
 
 	if (hs->sco) {
+		int sock = g_io_channel_unix_get_fd(hs->sco);
+		shutdown(sock, SHUT_RDWR);
 		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 		hs->sco = NULL;
@@ -1456,6 +1458,7 @@ failed_not_supported:
 	if (p->msg)
 		error_not_supported(dev->conn, p->msg);
 failed:
+	p->svclass = 0;
 	pending_connect_finalize(dev);
 	headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 }
@@ -2084,6 +2087,26 @@ void headset_update(struct audio_device *dev, uint16_t svc,
 	}
 }
 
+static int headset_close_rfcomm(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
+
+	if (rfcomm) {
+		g_io_channel_shutdown(rfcomm, TRUE, NULL);
+		g_io_channel_unref(rfcomm);
+		hs->tmp_rfcomm = NULL;
+		hs->rfcomm = NULL;
+	}
+
+	hs->data_start = 0;
+	hs->data_length = 0;
+
+	hs->nrec = TRUE;
+
+	return 0;
+}
+
 static void headset_free(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
@@ -2096,15 +2119,9 @@ static void headset_free(struct audio_device *dev)
 	if (hs->dc_id)
 		device_remove_disconnect_watch(dev->btd_dev, hs->dc_id);
 
-	if (hs->sco) {
-		g_io_channel_shutdown(hs->sco, TRUE, NULL);
-		g_io_channel_unref(hs->sco);
-	}
+	close_sco(dev);
 
-	if (hs->rfcomm) {
-		g_io_channel_shutdown(hs->rfcomm, TRUE, NULL);
-		g_io_channel_unref(hs->rfcomm);
-	}
+	headset_close_rfcomm(dev);
 
 	g_free(hs);
 	dev->headset = NULL;
@@ -2405,26 +2422,6 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 	return 0;
 }
 
-static int headset_close_rfcomm(struct audio_device *dev)
-{
-	struct headset *hs = dev->headset;
-	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
-
-	if (rfcomm) {
-		g_io_channel_shutdown(rfcomm, TRUE, NULL);
-		g_io_channel_unref(rfcomm);
-		hs->tmp_rfcomm = NULL;
-		hs->rfcomm = NULL;
-	}
-
-	hs->data_start = 0;
-	hs->data_length = 0;
-
-	hs->nrec = TRUE;
-
-	return 0;
-}
-
 static void disconnect_cb(struct btd_device *btd_dev, gboolean removal,
 				void *user_data)
 {
@@ -2766,7 +2763,7 @@ int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
 	if (!active_devices)
 		return -ENODEV;
 
-	if (number)
+	if (number && strlen(number) > 0)
 		send_foreach_headset(active_devices, hfp_cmp,
 				"\r\n+CLCC: %d,%d,%d,%d,%d,\"%s\",%d\r\n",
 				idx, dir, status, mode, mprty, number, type);
