@@ -41,10 +41,11 @@
 #include "btio.h"
 #include "plugin.h"
 #include "logging.h"
-#include "unix.h"
 #include "device.h"
+#include "unix.h"
 #include "headset.h"
 #include "manager.h"
+#include "gateway.h"
 
 static GIOChannel *sco_server = NULL;
 
@@ -55,6 +56,8 @@ static GKeyFile *load_config_file(const char *file)
 
 	keyfile = g_key_file_new();
 
+	g_key_file_set_list_separator(keyfile, ',');
+
 	if (!g_key_file_load_from_file(keyfile, file, 0, &err)) {
 		error("Parsing %s failed: %s", file, err->message);
 		g_error_free(err);
@@ -70,7 +73,7 @@ static void sco_server_cb(GIOChannel *chan, GError *err, gpointer data)
 	int sk;
 	struct audio_device *device;
 	char addr[18];
-	bdaddr_t dst;
+	bdaddr_t src, dst;
 
 	if (err) {
 		error("sco_server_cb: %s", err->message);
@@ -78,6 +81,7 @@ static void sco_server_cb(GIOChannel *chan, GError *err, gpointer data)
 	}
 
 	bt_io_get(chan, BT_IO_SCO, &err,
+			BT_IO_OPT_SOURCE_BDADDR, &src,
 			BT_IO_OPT_DEST_BDADDR, &dst,
 			BT_IO_OPT_DEST, addr,
 			BT_IO_OPT_INVALID);
@@ -86,27 +90,47 @@ static void sco_server_cb(GIOChannel *chan, GError *err, gpointer data)
 		goto drop;
 	}
 
-	device = manager_find_device(&dst, NULL, FALSE);
+	device = manager_find_device(NULL, &src, &dst, AUDIO_HEADSET_INTERFACE,
+					FALSE);
+	if (!device)
+		device = manager_find_device(NULL, &src, &dst,
+						AUDIO_GATEWAY_INTERFACE,
+						FALSE);
+
 	if (!device)
 		goto drop;
 
-	if (headset_get_state(device) < HEADSET_STATE_CONNECTED) {
-		debug("Refusing SCO from non-connected headset");
-		goto drop;
-	}
+	if (device->headset) {
+		if (headset_get_state(device) < HEADSET_STATE_CONNECTED) {
+			debug("Refusing SCO from non-connected headset");
+			goto drop;
+		}
 
-	if (!get_hfp_active(device)) {
-		error("Refusing non-HFP SCO connect attempt from %s", addr);
+		if (!get_hfp_active(device)) {
+			error("Refusing non-HFP SCO connect attempt from %s",
+									addr);
+			goto drop;
+		}
+
+		if (headset_connect_sco(device, chan) < 0)
+			goto drop;
+
+		headset_set_state(device, HEADSET_STATE_PLAYING);
+	} else if (device->gateway) {
+		if (!gateway_is_connected(device)) {
+			debug("Refusing SCO from non-connected AG");
+			goto drop;
+		}
+
+		if (gateway_connect_sco(device, chan) < 0)
+			goto drop;
+	} else
 		goto drop;
-	}
 
 	sk = g_io_channel_unix_get_fd(chan);
 	fcntl(sk, F_SETFL, 0);
 
-	if (headset_connect_sco(device, chan) == 0) {
-		debug("Accepted SCO connection from %s", addr);
-		headset_set_state(device, HEADSET_STATE_PLAYING);
-	}
+	debug("Accepted SCO connection from %s", addr);
 
 	return;
 
@@ -128,23 +152,32 @@ static int audio_init(void)
 
 	if (unix_init() < 0) {
 		error("Unable to setup unix socket");
-		return -EIO;
+		goto failed;
 	}
 
-	if (audio_manager_init(connection, config) < 0) {
-		dbus_connection_unref(connection);
-		return -EIO;
-	}
+	if (audio_manager_init(connection, config) < 0)
+		goto failed;
 
 	sco_server = bt_io_listen(BT_IO_SCO, sco_server_cb, NULL, NULL,
 					NULL, NULL,
 					BT_IO_OPT_INVALID);
 	if (!sco_server) {
 		error("Unable to start SCO server socket");
-		return -EIO;
+		goto failed;
 	}
 
 	return 0;
+
+failed:
+	audio_manager_exit();
+	unix_exit();
+
+	if (connection) {
+		dbus_connection_unref(connection);
+		connection = NULL;
+	}
+
+	return -EIO;
 }
 
 static void audio_exit(void)
@@ -162,4 +195,5 @@ static void audio_exit(void)
 	dbus_connection_unref(connection);
 }
 
-BLUETOOTH_PLUGIN_DEFINE("audio", audio_init, audio_exit)
+BLUETOOTH_PLUGIN_DEFINE(audio, VERSION,
+			BLUETOOTH_PLUGIN_PRIORITY_DEFAULT, audio_init, audio_exit)
