@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2000-2002  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2003-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2003-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -603,7 +603,7 @@ do_connect:
 	return;
 }
 
-static void inline hci_link_control(uint16_t ocf, int plen, uint8_t *data)
+static void hci_link_control(uint16_t ocf, int plen, uint8_t *data)
 {
 	uint8_t status;
 
@@ -632,7 +632,7 @@ static void inline hci_link_control(uint16_t ocf, int plen, uint8_t *data)
 	}
 }
 
-static void inline hci_link_policy(uint16_t ocf, int plen, uint8_t *data)
+static void hci_link_policy(uint16_t ocf, int plen, uint8_t *data)
 {
 	uint8_t status;
 
@@ -646,7 +646,7 @@ static void inline hci_link_policy(uint16_t ocf, int plen, uint8_t *data)
 	}
 }
 
-static void inline hci_host_control(uint16_t ocf, int plen, uint8_t *data)
+static void hci_host_control(uint16_t ocf, int plen, uint8_t *data)
 {
 	read_local_name_rp ln;
 	read_class_of_dev_rp cd;
@@ -745,7 +745,7 @@ static void inline hci_host_control(uint16_t ocf, int plen, uint8_t *data)
 	}
 }
 
-static void inline hci_info_param(uint16_t ocf, int plen, uint8_t *data)
+static void hci_info_param(uint16_t ocf, int plen, uint8_t *data)
 {
 	read_local_version_rp lv;
 	read_local_features_rp lf;
@@ -1314,7 +1314,8 @@ int main(int argc, char *argv[])
 	vdev.features[7] = 0x80;
 
 	memset(vdev.name, 0, sizeof(vdev.name));
-	strncpy((char *) vdev.name, "BlueZ (Virtual HCI)", sizeof(vdev.name));
+	strncpy((char *) vdev.name, "BlueZ (Virtual HCI)",
+							sizeof(vdev.name) - 1);
 
 	vdev.dev_class[0] = 0x00;
 	vdev.dev_class[1] = 0x00;
